#include "GLFW/glfw3.h"
#include "ui.h"
#include "gpu.h"
#include "draw.h"

#include "arpa/inet.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"
#include "pthread.h"

typedef struct ClientContextStruct {
  GLFWwindow* window;
  RenderContext render;
  UIContext ui;
} ClientContext;

void* network_thread(void* data) {
  ClientContext* context = (ClientContext*)data;
  (void)context;

  return NULL;
}

VkResult main_thread(ClientContext* context) {
  VkResult result;

  GPUString strings[] = {
    {
      .pos = {0, 32},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 0,
      .length = 0,
      .font = 0,
    },
    {
      .pos = {110, 138},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 20,
      .length = 1,
      .font = 0,
    },
    {
      .pos = {210, 138},
      .size = 32,
      .color = {1.0, 1.0, 1.0, 1.0},
      .offset = 21,
      .length = 1,
      .font = 0,
    },
  };

  GPUDrawable drawables[] = {
    {
      .pos = {100, 100},
      .size = {50, 50},
      .color = {1, 0, 0, 1},
      .type = 0,
      .id = 1,
    },
    {
      .pos = {200, 100},
      .size = {50, 50},
      .color = {0, 1, 0, 1},
      .type = 0,
      .id = 2,
    },
  };

  LayerInput layer = {
    .strings = strings,
    .num_strings = sizeof(strings)/sizeof(GPUString),
    .max_codes = 24,

    .drawables = drawables,
    .num_drawables = sizeof(drawables)/sizeof(GPUDrawable),
  };

  ContainerInput container = {
    .id = 1,
    .size = {WINDOW_MIN_WIDTH, WINDOW_MIN_HEIGHT},
    .layer_count = 1,
    .layers = &layer,
  };

  create_container(&container, &context->render, &context->ui);
  map_string("1234", context->ui.containers[0].layers[0].codes_buffer, 20, 0, &context->ui);
  VK_RESULT(add_transfers(
        &context->ui.containers[0].layers[0].codes_buffer[20],
        context->ui.containers[0].layers[0].codes,
        20*sizeof(uint32_t),
        4*sizeof(uint32_t),
        &context->render));

  //
  uint32_t* mapped_codes = context->ui.containers[0].layers[0].codes_buffer;
  GPUString* mapped_string = context->ui.containers[0].layers[0].strings_buffer;
  char str[21] = "1234567890";
  uint32_t last_id = 0;
  uint32_t last_last_id = -1;

  int test = 0;
  //

  while(glfwWindowShouldClose(context->window) == 0) {
    glfwPollEvents();
    double frame_time = glfwGetTime();

    if(last_id != last_last_id) {
      last_last_id = last_id;
      test = (test + 1) % 7;

      snprintf(str, 21, "Last ID: %d", last_id);
      map_string(str, mapped_codes, 0, 0, &context->ui);
      VK_RESULT(add_transfers(
            context->ui.containers[0].layers[0].codes_buffer,
            context->ui.containers[0].layers[0].codes,
            0,
            strlen(str)*sizeof(uint32_t),
            &context->render));
      mapped_string->length = strlen(str);
      VK_RESULT(add_transfers(
            &context->ui.containers[0].layers[0].strings_buffer[0].length,
            context->ui.containers[0].layers[0].strings,
            offsetof(GPUString, length),
            sizeof(uint32_t),
            &context->render));

    }
    //

    VkResult result = draw_frame(&context->render, &context->ui, frame_time);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      glfwDestroyWindow(context->window);
    }

  }

  return 0;
}

void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
  (void)window;
  (void)key;
  (void)scancode;
  (void)action;
  (void)mods;
}

bool contains(double* point, GPUContainer* container, GPUDrawable* rect) {
  vec2 pos = {
    container->offset[0] + rect->pos[0],
    container->offset[1] + rect->pos[1],
  };
  return (point[0] >= pos[0] && point[0] <= pos[0] + rect->size[0]) &&
         (point[1] >= pos[1] && point[1] <= pos[1] + rect->size[1]) &&
         (point[0] >= container->offset[0] && point[0] <= container->offset[0] + container->size[0]) &&
         (point[1] >= container->offset[1] && point[1] <= container->offset[1] + container->size[1]);
}

void mouse_button_callback(GLFWwindow* window, int button, int action, int mods) {
  ClientContext* context = (ClientContext*)glfwGetWindowUserPointer(window);

  (void)mods;
  (void)context;
  double cursor[2];
  glfwGetCursorPos(window, &cursor[0], &cursor[1]);
  switch(button) {
  // Handle camera hover
  case GLFW_MOUSE_BUTTON_MIDDLE:
    if(action == GLFW_PRESS) {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
    } else {
      glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }
    break;
  case GLFW_MOUSE_BUTTON_LEFT:
    if(action == GLFW_PRESS) {
      for(uint32_t c = context->ui.max_containers - 1; c >= 0; c--) {
        if(context->ui.containers[c].id == 0x00000000) {
          continue;
        }

        for(uint32_t l = 0; l < context->ui.containers[c].layer_count; l++) {
          for(uint32_t d = 0; d < context->ui.containers[c].layers[l].data.num_drawables; d++) {
            if(contains(cursor, &context->ui.containers[c].data, &context->ui.containers[c].layers[l].drawables_buffer[d])) {
              fprintf(stderr, "hit\n");
            }
          }
        }
      }
      // NOTE: this logic is the same as it would be for right-click(just with different outcomes), so dont repeat yourself
      // 1. Search through the UI context for the first element that the contains the mouse point
      // 2. If no UI element intersection, cast a ray through the world scene to find the "clicked entity" 
      // 3. Based on what was clicked, start the mouse click animation at the target area
    }
    break;
  case GLFW_MOUSE_BUTTON_RIGHT:
    if(action == GLFW_PRESS) {
    }
    break;
  }
}

void scroll_callback(GLFWwindow* window, double xoffset, double yoffset) {
  (void)window;
  (void)xoffset;
  (void)yoffset;
}

void cursor_pos_callback(GLFWwindow* window, double xpos, double ypos) {
  (void)window;
  (void)xpos;
  (void)ypos;
}

int main() {
  ClientContext context = {};
  context.window = init_window();
  if(context.window == NULL) {
    return 1;
  }

  glfwSetWindowUserPointer(context.window, &context);
  glfwSetKeyCallback(context.window, key_callback);
  glfwSetMouseButtonCallback(context.window, mouse_button_callback);
  glfwSetScrollCallback(context.window, scroll_callback);
  glfwSetCursorPosCallback(context.window, cursor_pos_callback);

  int error;
  VkResult result;
  VK_RESULT(init_vulkan(context.window, &context.render));
  
  // TODO: make # of fonts/textures/containers scaling, recreate GPU buffers as necessary
  VK_RESULT(create_ui_context(10, 10, 10, &context.render, &context.ui));

  pthread_t network_thread_handle;

  error = pthread_create(&network_thread_handle, NULL, &network_thread, &context);
  if(error != 0) {
    return error;
  }

  error = main_thread(&context);
  if(error != 0) {
    return error;
  }

  error = pthread_join(network_thread_handle, NULL);
  if(error != 0) {
    return error;
  }

  return 0;
}
